package org.processmining.v7.bruteforce;

import java.util.ArrayList;
import java.util.Collection;

import org.apache.commons.math3.optim.PointValuePair;
import org.apache.commons.math3.optim.linear.LinearConstraint;
import org.apache.commons.math3.optim.linear.LinearConstraintSet;
import org.apache.commons.math3.optim.linear.LinearObjectiveFunction;
import org.apache.commons.math3.optim.linear.LinearOptimizer;
import org.apache.commons.math3.optim.linear.Relationship;
import org.apache.commons.math3.optim.linear.SimplexSolver;

class ImplicitnessTester {
	public ImplicitnessTester() {
	}

	//Use an LPP to find implicit places____________________________________________________________________________________________
	public MyProcessModel testByLPP2(MyProcessModel pM) {
		//compute all the stuff needed for the LPP
		ArrayList<MyPlace> places = (ArrayList<MyPlace>) pM.getPlaces().clone();
		final ArrayList<MyPlace> finalPlaces = new ArrayList<MyPlace>();
		final String[] transitions = pM.getTransitions();
		//compute pre-incidence, post-incidence and incidence matrix of pM: for each place there is a row, for each transition a column
		final ArrayList<Integer[]> preIncMatrix = new ArrayList<Integer[]>();
		final ArrayList<Integer[]> postIncMatrix = new ArrayList<Integer[]>();
		final ArrayList<Integer[]> incMatrix = new ArrayList<Integer[]>();
		for (int p = 0; p < places.size(); p++) {
			preIncMatrix.add(new Integer[transitions.length]);
			postIncMatrix.add(new Integer[transitions.length]);
			incMatrix.add(new Integer[transitions.length]);
			MyPlace place = places.get(p);
			int pIn = place.getInputTrKey();
			int pOut = place.getOutputTrKey();
			for (int t = 0; t < transitions.length; t++) {
				if ((pIn & getMask(t, transitions)) > 0) {//t'th transition is contained in input of p
					preIncMatrix.get(p)[t] = 1;
				} else {
					preIncMatrix.get(p)[t] = 0;
				}
				if ((pOut & getMask(t, transitions)) > 0) {//t'th transition is contained in output of p
					postIncMatrix.get(p)[t] = 1;
				} else {
					postIncMatrix.get(p)[t] = 0;
				}
				incMatrix.get(p)[t] = preIncMatrix.get(p)[t] - postIncMatrix.get(p)[t];
			}
		}
		//do the LPP magic to check implicitness for each place
		//increase speed by removing implicit places for the next iteration
		ArrayList<MyPlace> tempPlaces = (ArrayList<MyPlace>) places.clone();
		while (!tempPlaces.isEmpty()) {
			int placePos = places.indexOf(tempPlaces.get(0));
			if (isImplicitByLPP2(placePos, places, transitions, preIncMatrix, incMatrix)) {
				//this place is implicit and can be deleted
				places.remove(placePos);
				preIncMatrix.remove(placePos);
				incMatrix.remove(placePos);
				tempPlaces.remove(0);
			} else {
				tempPlaces.remove(0);
				finalPlaces.add(places.get(placePos));
			}
		}
		
/*		//do the LPP magic to check implicitness for each place
		//DO NOT increase speed by removing implicit places for the next iteration
		for( MyPlace p : places) {
			int placePos = places.indexOf(p);
			if (!(isImplicitByLPP2(placePos, places, transitions, preIncMatrix, incMatrix))) {
				finalPlaces.add(places.get(placePos));
			}
		}
		*/
		
		pM.setPlaces(finalPlaces);
		return pM;
	}


	//use an LPP solver on the given parameters to test the place defined by placePos for implicitness
	//variables: y | z | k | x --> places.size()+places.size()+1+1
	private boolean isImplicitByLPP2(int currP, ArrayList<MyPlace> places, String[] transitions,
			ArrayList<Integer[]> preIncMatrix, ArrayList<Integer[]> incMatrix) {
		//For initial marking 0, variables k, x and reference sets Y, Z the objective functions is
		//0*y1+0*y2+ ... 0*yn + 0*z1+0*z2+ ... 0*zn + 1*k + 0*x + 0 
		//This simplifies to 1*k
		double[] coefficientsLinearObjectiveFunction = new double[(places.size() * 2) + 2]; //there are 2*|places| coefficients for Y, Z and 2 coefficient for k, x
		for (int i = 0; i < (places.size() * 2); i++) {
			coefficientsLinearObjectiveFunction[i] = 0;
		}
		coefficientsLinearObjectiveFunction[places.size() * 2] = 1; //k*1
		LinearObjectiveFunction objectiveFunction = new LinearObjectiveFunction(coefficientsLinearObjectiveFunction, 0);

		//Add the linear constraints w.r.t. the current place currP, using k, x, Y, Z, incMatrix, preIncMatrix
		Collection<LinearConstraint> constraints = new ArrayList<LinearConstraint>();
		//Type 0: ensure currP is not in Y, Z, that is currP=0
		double[] removeCurrPY = new double[(places.size() * 2) + 2];//coefficients are all 0, except for currP in Y
		double[] removeCurrPZ = new double[(places.size() * 2) + 2];//coefficients are all 0, except for currP in Z
		for (int i = 0; i < ((places.size() * 2) + 2); i++) {
			removeCurrPY[i] = 0;
			removeCurrPZ[i] = 0;
		}
		removeCurrPY[currP] = 1;
		removeCurrPZ[currP + places.size()] = 1;
		constraints.add(new LinearConstraint(removeCurrPY, Relationship.EQ, 0));
		constraints.add(new LinearConstraint(removeCurrPZ, Relationship.EQ, 0));
		//Type 1: Y>=Z>=0, k>=0, x=0, x<k--> x-k<=-0
		//Y>=Z>=0 --> Z>=0 AND Y-Z>=0
		for (int p = 0; p < places.size(); p++) {
			double[] nonNegZ = new double[(places.size() * 2) + 2];//coefficients are all 0, except for the current z=1
			double[] YGegZ = new double[(places.size() * 2) + 2];//coefficients are all 0, except for the current y=1 and z=-1
			for (int i = 0; i < ((places.size() * 2) + 2); i++) {//put everything to 0 by default
				nonNegZ[i] = 0;
				YGegZ[i] = 0;
			}
			nonNegZ[places.size() + p] = 1; //z=1
			YGegZ[p] = 1; //y=1
			YGegZ[places.size() + p] = -1; //z=-1
			constraints.add(new LinearConstraint(nonNegZ, Relationship.GEQ, 0));//1*z>=0	
			constraints.add(new LinearConstraint(YGegZ, Relationship.GEQ, 0));//1*y-1*z>=0	
		}
		//k>=0, x=0, x-k<=-1
		double[] nonNegk = new double[(places.size() * 2) + 2];//coefficients are all 0, except for k=1
		double[] zerox = new double[(places.size() * 2) + 2];//coefficients are all 0, except for x=1
		double[] xSmallerk = new double[(places.size() * 2) + 2];//coefficients are all 0, except for x=1, k=-1
		for (int i = 0; i < ((places.size() * 2) + 2); i++) {
			nonNegk[i] = 0;
			zerox[i] = 0;
		}
		nonNegk[places.size() * 2] = 1;
		zerox[(places.size() * 2) + 1] = 1;
		xSmallerk[places.size() * 2] = -1; //k=-1
		xSmallerk[(places.size() * 2) + 1] = 1; //x=1
		constraints.add(new LinearConstraint(xSmallerk, Relationship.LEQ, -1));//1*x-1*k <=-1	
		constraints.add(new LinearConstraint(nonNegk, Relationship.GEQ, 0));//1*k>=0	
		constraints.add(new LinearConstraint(zerox, Relationship.EQ, 0));//1*x=0	

		//Type 2: Y*incMatrix<=k*inc(currP) ---> Y*incMatrix - k*inc(currP) <=0;
		for (int t = 0; t < transitions.length; t++) {//add one constraint for each transition
			double[] coefficients = new double[(places.size() * 2) + 2];//coefficients are based on incMatrix
			for (int p = 0; p < places.size(); p++) {
				coefficients[p] = incMatrix.get(p)[t]; //find the coefficient for Y each pair (p,t)
				coefficients[p + places.size()] = 0; //Z is put to 0
			}
			//coefficient for k is -incMatrix[currP, t]
			coefficients[places.size() * 2] = incMatrix.get(currP)[t] * (-1);
			//coefficient for x is 0;
			coefficients[(places.size() * 2) + 1] = 0;
			//the sum should be <= inc(currP,t)
			constraints.add(new LinearConstraint(coefficients, Relationship.LEQ, 0));
		}

		//Type 3: forall t with currP in pre(t): Z*pre(q, t) + x >= k *pre(currP, t), for q in P/{currP}
		// --> Z*pre(q, t) + x - k* pre(currP, t) >=0 	//TODO in contrast to paper (imp places in net systems, garcia&colom, proposition 13) seems to work fine for k=1. explain this result theoretically?!
		for (int t = 0; t < transitions.length; t++) {
			if (preIncMatrix.get(currP)[t] == 1) {//for all t with currP in pre(t)
				double[] coefficients = new double[(places.size() * 2) + 2];
				for (int p = 0; p < places.size(); p++) {
					coefficients[p] = 0; //coefficients of Y=0
					coefficients[p + places.size()] = preIncMatrix.get(p)[t]; //coefficients of Z = pre(p,t)
				}
				//coefficients[places.size()*2] = (-1)*preIncMatrix.get(currP)[t];//coefficient of k=-pre(currP, t)
				coefficients[(places.size() * 2) + 1] = 1;//coefficient of x=1
				constraints.add(new LinearConstraint(coefficients, Relationship.GEQ, preIncMatrix.get(currP)[t]));
			}
		}

		LinearOptimizer solver = new SimplexSolver();
		try {
			solver.optimize(objectiveFunction, new LinearConstraintSet(constraints));
		} catch (org.apache.commons.math3.optim.linear.NoFeasibleSolutionException NFSE) {
			//				System.out.println("LPP Solver found no feasable solution for reference set of place "+placeToNamedString(places[currP], transitions));
			return false;
		}
		/*
		 * //for debuggin print out the found solution double[] solutionSupport
		 * = solution.getPoint(); String refSetStringY = "Y: "; String
		 * refSetStringZ = "Z: "; String k = "k: "; String x = "x: "; for (int
		 * pos = 0; pos < solutionSupport.length-2; pos++) { if
		 * (solutionSupport[pos] > 0 && pos < places.size()) { refSetStringY =
		 * refSetStringY + placeToNamedString(places.get(pos), transitions); }
		 * else if(solutionSupport[pos] > 0 && pos < places.size()*2){
		 * refSetStringZ = refSetStringZ +
		 * placeToNamedString(places.get(pos-places.size()), transitions); } } k
		 * = k+solutionSupport[places.size()*2]; x =
		 * x+solutionSupport[places.size()*2+1];
		 * System.out.println("Place "+placeToNamedString(places.get(currP),
		 * transitions)+" is implicit with solution ");
		 * System.out.println(refSetStringY); System.out.println(refSetStringZ);
		 * System.out.println(k); System.out.println(x);
		 */
		return true;
	}

	
	
	
	
	
	
	
		
	
	
	//____"Official" LPP to find implicit places (deprecated, update/check before use)_________________________________________________________
	public MyProcessModel testByLPP(final MyProcessModel pM) {
		//compute all the stuff needed for the LPP
		final ArrayList<MyPlace> finalPlaces = new ArrayList<MyPlace>();
		final String[] transitions = pM.getTransitions();
		ArrayList<MyPlace> places = (ArrayList<MyPlace>) pM.getPlaces().clone();

		//TO DO change arrays to arraylist and remove implicit places from it for each call!(see version 2)

		//compute pre-incidence, post-incidence and incidence matrix of pM: for each place there is a row, for each transition a column
		final ArrayList<Integer[]> preIncMatrix = new ArrayList<Integer[]>();
		final ArrayList<Integer[]> postIncMatrix = new ArrayList<Integer[]>();
		final ArrayList<Integer[]> incMatrix = new ArrayList<Integer[]>();
		for (int p = 0; p < places.size(); p++) {
			preIncMatrix.add(new Integer[transitions.length]);
			postIncMatrix.add(new Integer[transitions.length]);
			incMatrix.add(new Integer[transitions.length]);
			MyPlace place = places.get(p);
			int pIn = place.getInputTrKey();
			int pOut = place.getOutputTrKey();
			for (int t = 0; t < transitions.length; t++) {
				if ((pIn & getMask(t, transitions)) > 0) {//t'th transition is contained in input of p
					preIncMatrix.get(p)[t] = 1;
				} else {
					preIncMatrix.get(p)[t] = 0;
				}
				if ((pOut & getMask(t, transitions)) > 0) {//t'th transition is contained in output of p
					postIncMatrix.get(p)[t] = 1;
				} else {
					postIncMatrix.get(p)[t] = 0;
				}
				incMatrix.get(p)[t] = preIncMatrix.get(p)[t] - postIncMatrix.get(p)[t];
			}
		}

		/*
		 * String transitionString = "Transitions Order: "; for(String
		 * s:transitions) { transitionString=transitionString+s+", "; }
		 * System.out.println(transitionString);
		 */
		//for each place run the LPP-Solver to check for implicitness
		for (int placePos = 0; placePos < places.size(); placePos++) {
			//do the LPP magic for each place
			if (isImplicitByLPP(placePos, places, transitions, preIncMatrix, incMatrix)) {
				//this place can be deleted
			} else {
				finalPlaces.add(places.get(placePos));
			}
		}

		pM.setPlaces(finalPlaces);
		return pM;
	}

	//use an LPP solver on the given parameters to test the place defined by placePos for implicitness
	private boolean isImplicitByLPP(int currP, ArrayList<MyPlace> places, String[] transitions,
			ArrayList<Integer[]> preIncMatrix, ArrayList<Integer[]> incMatrix) {
		//For initial marking M, variable x and reference set Y the objective functions is
		//m1*y1+m2*y2+ ... mn*yn+1*x + 0
		//The empty ininital marking in our case simplifies this to 1*x + 0
		double[] coefficientsLinearObjectiveFunction = new double[places.size() + 1]; //there are |places| coefficients for Y and one coefficient for x
		for (int i = 0; i < places.size(); i++) {
			coefficientsLinearObjectiveFunction[i] = 0;
		}
		coefficientsLinearObjectiveFunction[places.size()] = 1;
		LinearObjectiveFunction objectiveFunction = new LinearObjectiveFunction(coefficientsLinearObjectiveFunction, 0);

		//Add the linear constraints w.r.t. the current place currP, using x, Y, incMatrix, preIncMatrix
		Collection<LinearConstraint> constraints = new ArrayList<LinearConstraint>();
		//Type 0: ensure currP is not in reference set, that is currP=0
		double[] removeCurrP = new double[places.size() + 1];//coefficients are all 0, except for currP
		for (int i = 0; i < removeCurrP.length; i++) {
			removeCurrP[i] = 0;
		}
		removeCurrP[currP] = 1;
		constraints.add(new LinearConstraint(removeCurrP, Relationship.EQ, 0));//1*currP=0
		//Type 1.1: Y>=0
		for (int p = 0; p < places.size(); p++) {
			double[] nonNegY = new double[places.size() + 1];//coefficients are all 0, except for the current y
			for (int i = 0; i < nonNegY.length; i++) {
				nonNegY[i] = 0;
			}
			nonNegY[p] = 1;
			constraints.add(new LinearConstraint(nonNegY, Relationship.GEQ, 0));//1*y>=0
		}
		//Type 1.2: Y<=1 TO DO necessary?
		for (int p = 0; p < places.size(); p++) {
			double[] singleOcc = new double[places.size() + 1];//coefficients are all 0, except for the current y
			for (int i = 0; i < singleOcc.length; i++) {
				singleOcc[i] = 0;
			}
			singleOcc[p] = 1;
			constraints.add(new LinearConstraint(singleOcc, Relationship.LEQ, 1));//1*y>=0
		}
		//Type 2: Y*incMatrix<=inc(currP)
		for (int t = 0; t < transitions.length; t++) {//add one constraint for each transition
			double[] coefficients = new double[places.size() + 1];//coefficients are based on incMatrix
			for (int p = 0; p < places.size(); p++) {
				coefficients[p] = incMatrix.get(p)[t]; //find the coefficient for each pair (p,t)
			}
			//coefficient for x is 0;
			coefficients[places.size()] = 0;
			//the sum should be <= inc(currP,t)
			constraints.add(new LinearConstraint(coefficients, Relationship.LEQ, incMatrix.get(currP)[t]));
		}
		//Type 3: forall t with currP in pre(t): pre(currP, t) - Y*preIncMatrix[t]<= x
		for (int t = 0; t < transitions.length; t++) {
			if (preIncMatrix.get(currP)[t] == 1) {//for all t with currP in pre(t)
				double[] coefficients = new double[places.size() + 1];
				for (int p = 0; p < places.size(); p++) {//coefficients of Y
					coefficients[p] = preIncMatrix.get(p)[t];
				}
				coefficients[places.size()] = 1;//coefficient of x
				constraints.add(new LinearConstraint(coefficients, Relationship.GEQ, preIncMatrix.get(currP)[t]));
			}
		}

		//create and run solver
		PointValuePair solution = null;
		LinearOptimizer solver = new SimplexSolver();
		try {
			solution = solver.optimize(objectiveFunction, new LinearConstraintSet(constraints));//default goal is minimize
		} catch (org.apache.commons.math3.optim.linear.NoFeasibleSolutionException NFSE) {
			//				System.out.println("LPP Solver found no feasable solution for reference set of place "+placeToNamedString(places[currP], transitions));
			return false;
		}

		if (solution.getValue() <= 0) {
			double[] solutionSupport = solution.getPoint();
			String refSetString = "";
			for (int pos = 0; pos < (solutionSupport.length - 1); pos++) {
				if (solutionSupport[pos] > 0) {
					refSetString = refSetString + placeToNamedString(places.get(pos), transitions);
				}
			}
			//				System.out.println("Place "+placeToNamedString(places[currP], transitions)+" is implicit with reference set "+refSetString);
			return true;
		}
		System.out.println("Feasable Solution for place " + placeToNamedString(places.get(currP), transitions)
				+ " might require initial marking.");
		return false; //TO DO bordercase: the LPP returns an upper bound for the necessary initial marking, thus it is possible that a result larger than 0 is not optimal
	}

	
	
	
	
	
	//Find and remove implicit places (old, very suboptimal solution)_______________________________________________________________________
	public MyProcessModel testForImplicitPlaces(final MyProcessModel pM) {
		MyProcessModel newPM = new MyProcessModel(new ArrayList<MyPlace>(), pM.getTransitions());
		for (MyPlace p : pM.getPlaces()) {
			if (isImplicit(p, pM)) {
			} else {
//				System.out.println("Is not implicit: (" + getTransitions(p.getInputTrKey(), pM) + "|"+ getTransitions(p.getOutputTrKey(), pM) + ")");
				newPM.addPlace(p);
			}
		}
		return newPM;
	}

	//Test an internal place p for being structurally implicit. p is implicit if there exists a set of places I in P\{p},
	//which is a reference set (refSet) for p in pM
	boolean isImplicit(final MyPlace p, final MyProcessModel pM) {
		boolean cutImpTest = false;
		ArrayList<MyPlace> testSet = new ArrayList<MyPlace>(pM.getPlaces()); //try to find a set I subset of testSet
		MyPlace start = null;
		MyPlace end = null;
		for (MyPlace p1 : testSet) {
			if (p1.getInputTrKey() == 0) {
				start = p1;
			}
			if (p1.getOutputTrKey() == 0) {
				end = p1;
			}
		}
		testSet.remove(p);
		testSet.remove(end);
		testSet.remove(start);
		//check every small subset of the testset until a reference set is found
		int maxNumber = (int) Math.min(Integer.MAX_VALUE, Math.pow(2, testSet.size())); //does not allow for int overflow
		for (int testSetKey = 2; testSetKey < maxNumber; testSetKey++) {
			if (Integer.bitCount(testSetKey) > 25) {//very bad solution to keep time short
				cutImpTest = true;
			} else {
				if (Integer.bitCount(testSetKey) > 1) {//refsets of size 0 are for contsant places, of size one for selfloops

					ArrayList<MyPlace> refSet = new ArrayList<MyPlace>();
					for (int i = 0; i < testSet.size(); i++) {
						if ((testSetKey & (1 << i)) > 0) {
							refSet.add(testSet.get(i));
						}
					}
					//test current I for being reference set for p
					if (isReferenceSet(p, refSet, pM.getTransitions())) {
						String debugstr = "Place (" + getTransitions(p.getInputTrKey(), pM) + "|"
								+ getTransitions(p.getOutputTrKey(), pM) + ") is implicit with reference set ";
						for (MyPlace p1 : refSet) {
							debugstr = debugstr + "(" + getTransitions(p1.getInputTrKey(), pM) + "|"
									+ getTransitions(p1.getOutputTrKey(), pM) + "), ";
						}
	//					System.out.println(debugstr);
						return true;
					}
				}
			}
		}
		if (cutImpTest) {
		}
		return false;
	}

	//test whether I is a reference set for p in pM, that is for every transition t:
	// Condition 1:  |(t,p)|-|(p,t)| >= SUM(|(t,q)|-|(t,q')|), with q,q' in I --> when t fires the increase of tokens in p is at least as big as the increase of tokens in I
	//Condition 2: |(p,t)| <= SUM(|(q,t)|), with q in I
	private boolean isReferenceSet(final MyPlace candidatePlace, final ArrayList<MyPlace> candidateI,
			final String[] transitions) {
		for (int i = 0; i < transitions.length; i++) { //test conditions for all transitions
			if (!checkCond2(candidatePlace, candidateI, transitions, i)) {//check condition 2
				return false;
			}
			if (!checkCond1(candidatePlace, candidateI, transitions, i)) {//check condition 1
				return false;
			}
		}
		//conditions hold for all transitions
		return true;
	}

	// Condition 1:  |(t,p)|-|(p,t)| >= SUM(|(t,q)|-|(t,q')|), with q,q' in I --> when t fires the increase of tokens in p is at least as big as the increase of tokens in I
	private boolean checkCond1(final MyPlace candidatePlace, final ArrayList<MyPlace> candidateI,
			final String[] transitions, final int transition) {
		MyPlace p = candidatePlace.copy();
		ArrayList<MyPlace> I = new ArrayList<MyPlace>(candidateI);
		int consumedTokensI = 0;
		int consumeP = 0;
		int addedTokensI = 0;
		int addP = 0;
		for (MyPlace otherP : I) {
			if ((otherP.getInputTrKey() & getMask(transition, transitions)) > 0) {//otherP contains transition as input
				addedTokensI++;
			}
			if ((otherP.getOutputTrKey() & getMask(transition, transitions)) > 0) {//otherP contains transition as output
				consumedTokensI++;
			}
		}
		if ((p.getInputTrKey() & getMask(transition, transitions)) > 0) {//p contains transition as input
			addP++;
		}
		if ((p.getOutputTrKey() & getMask(transition, transitions)) > 0) {//p contains transition as output
			consumeP++;
		}
		//actually check the condition
		if ((addP - consumeP) >= (addedTokensI - consumedTokensI)) {//check condition 1 for this transition
			return true;
		} else {
			return false;
		}
	}

	//Condition 2: |(p,t)| <= SUM(|(q,t)|), with q in I
	private boolean checkCond2(final MyPlace candidatePlace, final ArrayList<MyPlace> candidateI,
			final String[] transitions, final int transition) {
		if ((getMask(transition, transitions) & candidatePlace.getInputTrKey()) > 0) {//if the transition isn't part of the candidate places input transitions, condition 2 holds
			for (MyPlace otherP : candidateI) {//if no q in I contains the transition as input return false
				if ((getMask(transition, transitions) & otherP.getInputTrKey()) > 0) {
					return true;
				}
			}
			return false;
		} else {
			return true;
		}
	}

	//Utility Methods_______________________________________________________________________________________________

	private String placeToNamedString(final MyPlace p, final String[] transitions) {
		return getTransitionNames(p.getInputTrKey(), transitions).toString() + "|"
				+ getTransitionNames(p.getOutputTrKey(), transitions).toString() + ",";
	}

	//returns a collection containing all transitions names from the given transitions array
	private Collection<String> getTransitionNames(final int key, final String[] transitions) {
		Collection<String> result = new ArrayList<String>();
		if (key > (Math.pow(2, transitions.length))) {
			return null;
		}
		for (int i = 0; i < transitions.length; i++) {
			if ((key & getMask(i, transitions)) > 0) { //test key for ones
				result.add(transitions[i]);
			}
		}

		return result;
	}

	//return the transitions corresponding to the given key
	private Collection<String> getTransitions(final int key, final MyProcessModel pM) {
		Collection<String> result = new ArrayList<String>();
		if (key > Math.pow(2, pM.getTransitions().length)) {
			return null;
		}
		for (int i = 0; i < pM.getTransitions().length; i++) {
			if ((key & getMask(i, pM.getTransitions())) > 0) { //test key for ones
				result.add(pM.getTransitions()[i]);
			}
		}
		return result;
	}

	//return bitmask corresponding to position in the transition array
	private int getMask(final int position, final String[] transitions) {
		return (1 << (transitions.length - 1 - position));
	}

	public boolean isImplicit2(MyPlace currentP, MyProcessModel pM) {
		//compute all the stuff needed for the LPP
		ArrayList<MyPlace> places = (ArrayList<MyPlace>) pM.getPlaces().clone();
		places.add(currentP);
		final String[] transitions = pM.getTransitions();
		//compute pre-incidence, post-incidence and incidence matrix of pM: for each place there is a row, for each transition a column
		final ArrayList<Integer[]> preIncMatrix = new ArrayList<Integer[]>();
		final ArrayList<Integer[]> postIncMatrix = new ArrayList<Integer[]>();
		final ArrayList<Integer[]> incMatrix = new ArrayList<Integer[]>();
		for (int p = 0; p < places.size(); p++) {
			preIncMatrix.add(new Integer[transitions.length]);
			postIncMatrix.add(new Integer[transitions.length]);
			incMatrix.add(new Integer[transitions.length]);
			MyPlace place = places.get(p);
			int pIn = place.getInputTrKey();
			int pOut = place.getOutputTrKey();
			for (int t = 0; t < transitions.length; t++) {
				if ((pIn & getMask(t, transitions)) > 0) {//t'th transition is contained in input of p
					preIncMatrix.get(p)[t] = 1;
				} else {
					preIncMatrix.get(p)[t] = 0;
				}
				if ((pOut & getMask(t, transitions)) > 0) {//t'th transition is contained in output of p
					postIncMatrix.get(p)[t] = 1;
				} else {
					postIncMatrix.get(p)[t] = 0;
				}
				incMatrix.get(p)[t] = preIncMatrix.get(p)[t] - postIncMatrix.get(p)[t];
			}
		}
		//do the LPP magic to check implicitness for each place
		//increase speed by removing implicit places for the next iteration
		int placePos = places.indexOf(currentP);
		if (isImplicitByLPP2(placePos, places, transitions, preIncMatrix, incMatrix)) {
		//this place is implicit 
			return true;
		} else {
			return false;
		}
	}

}
