package org.processmining.v7.postproc_during_tc;

import java.util.ArrayList;

/*
 * This class public method getNext() returns the next candidate place based on
 * the given candidate place.
 */

//ASSUME THAT POS 0 OF TRANSITIONS IS END, POS 0 OF OUTMAPPING MAPS TO START
public class MyCandidates {
	private final String[] transitions;
	private final int[] outTrMapping;
	private final int largestOutIndex;
	private final int largestInIndex;
	private final int startMask; //a mask to find the bit corresponding to the startIndex in the transitions array
	private final int endMask;
	private final ArrayList<MyPlace> roots;
	private MyPlace currentRoot;

	public MyCandidates(final String[] transitions, final int startIndex, final int endIndex, final int[] outTrMapping,
		final Integer[] place_restrictions) {
		this.transitions = transitions;
		this.outTrMapping = outTrMapping;
		startMask = getMask(startIndex);
		endMask = getMask(endIndex);
		roots = computeBaseRoots();
		currentRoot = roots.get(0);
		this.largestOutIndex = getTransitionIndex(outTrMapping.length-1);
		this.largestInIndex = transitions.length-1;
	}

	//adds all base roots, that are places with only one in and one out transition (first level places)
	private ArrayList<MyPlace> computeBaseRoots() {
		ArrayList<MyPlace> baseRoots = new ArrayList<MyPlace>();
		for (int in = 1; in < transitions.length; in++) {
			for (int out = 1; out < transitions.length; out++) {
				baseRoots.add(new MyPlace(getMask(in), getMask(getTransitionIndex(out))));
			}
		}
		return baseRoots;
	}

	//returns the index of the transition array corresponding to the given index of the out transition order
	private int getTransitionIndex(int outIndex) {
		return outTrMapping[outIndex];
	}
	
	private int getHighOutIndex(int lastOutKey) {
		for (int i = outTrMapping.length - 1; i >= 0; i--) {
			if ((getMask(getTransitionIndex(i)) & lastOutKey) > 0) {
				return i;
			}
		}
		return 0;
	}
	
	private int getLowOutIndex(int lastOutKey) {
		for (int i = 0; i < outTrMapping.length; i++) {
			if ((getMask(getTransitionIndex(i)) & lastOutKey) > 0) {
				return i;
			}
		}
		return outTrMapping.length-1;
	}

	//returns the next place based on the given last place
	//uses a tree structure ordered according to transitions array (ingoing) and mapping (outgoing).
	//Makes use of the class variables roots and current root to keep track of tree traversal
	// Cuts off uninteresting branches based on fitness, other criteria planned
	public MyPlace getNext(MyPlace lastP, PlaceStatus fitness) {
		//case first place
		if (lastP == null) {
			if (isUseful(currentRoot.getInputTrKey(), currentRoot.getOutputTrKey())) {
				return currentRoot;
			} else {
				lastP = currentRoot;
				fitness = PlaceStatus.UNKNOWN;
			}
		}
		while (true) {
			MyPlace nextP = new MyPlace(0, 0);
			//Case 1: next place is descendant of last place
			nextP = findNextChild(lastP, fitness);
			//			System.out.println("Case 1: " + nextP.toBinaryString());
			//Case 2: next place contained in same subtree (based on currentRoot) but different branch
			if ((nextP.getInputTrKey() == 0) || (nextP.getOutputTrKey() == 0)) {
				nextP = findNextBranch(lastP);
			//					System.out.println("Case 2: " + nextP.toBinaryString());
			}
			//Case 3: next place contained in new subtree (new currentRoot needed)
			if ((nextP.getInputTrKey() == 0) || (nextP.getOutputTrKey() == 0)) {
				nextP = findNextRoot(lastP);
			//					System.out.println("Case 3: " + nextP.toBinaryString());
			}
			//Case 4: no more places to be found
			if ((nextP.getInputTrKey() == 0) || (nextP.getOutputTrKey() == 0)) {
			//					System.out.println("Case 4: " + nextP.toBinaryString());
				return null;
			}
			//return new place if useful, else keep searching
			if (isUseful(nextP.getInputTrKey(), nextP.getOutputTrKey())) {
				return nextP;
			}
			lastP = nextP;
			fitness = PlaceStatus.UNKNOWN;
		}
	}

	//try: find child of the last candidate node (deepen current branch), outchild first (NO backtracking)
	// possibly cut off uninteresting branches
	private MyPlace findNextChild(final MyPlace lastP, final PlaceStatus fitness) {
		int nextInKey = 0;
		int nextOutKey = 0;
		nextOutKey = findNextOutKey(lastP.getOutputTrKey(), fitness); //adds a new out-transitions to current candidate, returns 0 if impossible
		if (nextOutKey != 0) {
			nextInKey = lastP.getInputTrKey();
		} else {
			nextInKey = findNextInKey(lastP.getInputTrKey(), lastP.getOutputTrKey(), fitness); //add new in-transition to current candidate (only possible if outkey has exactly one transition), returns 0 if impossible
			if (nextInKey != 0) {
				nextOutKey = lastP.getOutputTrKey(); //should contain exactly one transition, otherwise inkey should be 0
			}
		}
		return new MyPlace(nextInKey, nextOutKey);
	}

	//adds a new out transition to current candidate (NO BACKTRACKING), possibly cuts of current branch, returns 0 on fail
	//current implementation: increase inkey from left to right, outkey from right to left
	private int findNextOutKey(final int lastOutKey, final PlaceStatus fitness) {
		int nextOutKey = 0;
		//if path is not at end AND last place was underfed --> cut this path
		if (((lastOutKey & getMask(largestOutIndex)) == 0) //last out transition is not added
				&& ((fitness == PlaceStatus.UNDERFED) || (fitness == PlaceStatus.MALFED))) {//previous place was underfed
		//	System.out.println("Cutting off underfed path!");
		}
		//if path is not cut, add output transition if possible (add only transitions larger than the largest current out transition)
		else if ((lastOutKey & getMask(largestOutIndex)) == 0) {//ensure largest transition isn't set
			int newOutTransition = getMask(getTransitionIndex((getHighOutIndex(lastOutKey) + 1)));//this puts a 1 to the position of the next transition
			nextOutKey = lastOutKey | newOutTransition;
		}
		return nextOutKey;
	}

	
	//adds a new in transition to current candidate (NO BACKTRACKING), possibly cuts of current branch, returns 0 on fail
	private int findNextInKey(final int lastInKey, final int lastOutKey, final PlaceStatus fitness) {
		int nextInKey = 0;
		// in transitions can only be added to places with exactly one output transition
		if (Integer.bitCount(lastOutKey) == 1) {
			//path is not at end AND no outkeys can be added, so overfedness of previous place can be exploited to cut branch
			if (((lastInKey & getMask(largestInIndex)) == 0) && (lastOutKey == getMask(largestOutIndex))
					&& ((fitness == PlaceStatus.OVERFED) || (fitness == PlaceStatus.MALFED))) {
			//	System.out.println("Cutting off overfed path!");
			}
			// case input can be added (add only transitions larger than the largest in transition)
			else if ((lastInKey & getMask(largestInIndex)) == 0) {//ensure largest transition isn't set
				int newInTransition = (Integer.lowestOneBit(lastInKey) >> 1);//this puts a 1 to the position of the next transition
				nextInKey = lastInKey | newInTransition;
			}
		}
		return nextInKey;
	}

	
	//discards subtree based on current root and starts traversing subtree based on next root
	private MyPlace findNextRoot(final MyPlace lastP) {
		roots.remove(currentRoot);
		if (!roots.isEmpty()) {
			currentRoot = roots.get(0);
			return currentRoot;
		} else {
			return new MyPlace(0, 0);
		}
	}


	
	//assume the current candidate cannot have children
	//try: backtrack at most up to current root, to find a new branch of this subtree
	private MyPlace findNextBranch(final MyPlace lastP) {
		int lastOutKey = lastP.getOutputTrKey();
		int lastInKey = lastP.getInputTrKey();
		int nextInKey = 0;
		int nextOutKey = 0;
		nextOutKey = backtrackOutKey(lastOutKey);
		if (nextOutKey != 0) {//new outbranch could be found
			nextInKey = lastInKey;
		} else {//no outbranch could be found
			lastOutKey = getMask(getTransitionIndex(getLowOutIndex(lastOutKey))); //ensure only the smallest out-transition is left ("root" of this outkey)
			//check whether the root of this outtree admits a new inbranch before backtracking
			nextInKey = findNextInKey(lastInKey, lastOutKey, PlaceStatus.UNKNOWN);
			if (nextInKey == 0) {//this outkey "root" doesn't admit any children
				nextInKey = backtrackInKey(lastInKey);
			}
			if (nextInKey != 0){//new inbranch could be found
				nextOutKey = lastOutKey;
			}
		}
		MyPlace nextP = new MyPlace(nextInKey, nextOutKey);
		if (isInCurrentSubtree(nextP)) {
			return nextP;
		} else {//no more places in current subtree
			return new MyPlace(0, 0);
		}
	}
	
	//returns true if the given place is a descendant of the currenRoot
	private boolean isInCurrentSubtree(MyPlace nextPlace) {
		if (isSubset(currentRoot.getInputTrKey(), nextPlace.getInputTrKey())
				&& isSubset(currentRoot.getOutputTrKey(), nextPlace.getOutputTrKey())) {
			return true;
		}
		return false;
	}

	//returns true if the set of transitions encoded by subset are all contained in the set of transitions encoded by superset
	private boolean isSubset(int subset, int superset) {
		if ((subset & superset) == subset) {//subset is completely contained in superset (possibly equal)
			return true;
		}
		return false;
	}
	
	//based on the last outkey, backtracks the outbranch until a new outbranch is found, returns 0 on fail
	//assume no children can be added to current candidate
	private int backtrackOutKey(int lastOutKey) {
		int nextOutKey = 0;
		if (Integer.bitCount(lastOutKey) > 1) {//if there is only one out transition, backtracking is impossible
			int lastOutIndex = getHighOutIndex(lastOutKey);//find the largest remaining out transition of the last candidate
			lastOutKey = getMask(getTransitionIndex(lastOutIndex)) ^ lastOutKey;//remove this largest transition to get back to candidates' parent
			// Case 1: the last branch was cut, this parent has another outbranch (last outbranch is always exactly one candidate and cannot be cut)
			if (getTransitionIndex(lastOutIndex) != largestOutIndex) {
				nextOutKey = lastOutKey | getMask(getTransitionIndex(lastOutIndex+1)); //find next outbranch of the parent
			}
			//case 2: the last branch was at it's end, so this parent cannot have other out branches
			//--> try to find the grandparent (parent of this parent) and find a new out branch of this grandparent
			else if ((nextOutKey == 0) && (Integer.bitCount(lastOutKey) > 1)) { //if this parent has only one out transition, further backtracking is impossible
				lastOutIndex = getHighOutIndex(lastOutKey);//find the largest remaining out transition of the last candidate
				lastOutKey = getMask(getTransitionIndex(lastOutIndex)) ^ lastOutKey;//remove this transition to get back to candidates' grandparent
				nextOutKey = lastOutKey | getMask(getTransitionIndex(lastOutIndex+1)); //add the next output transition to grandparent (should always be possible)
			}
		}
		return nextOutKey;
	}
	
	
	//based on the last inkey, backtracks the inbranch until new inbranch is found, returns 0 on fail
	//assume no children can be added to current candidate
	private int backtrackInKey(int lastInKey) {
		int nextInKey = 0;
		if (Integer.bitCount(lastInKey) > 1) {//if there is only one intransition backtracking is impossible
			int lastTransition = Integer.lowestOneBit(lastInKey); //find the largest remaining in transition of the last candidate
			lastInKey = lastTransition ^ lastInKey;//remove this largest transition to get back to candidates' parent
			//Case 1: last inbranch was cut, find next inbranch of this parent (should always be possible)
			if (lastTransition != 1) {
				nextInKey = lastInKey | (lastTransition >> 1);
			}
			//Case 2: this inbranch was at it's end, parent cannot have other inbranch
			//--> find parent of this parent (grandparent of this candidate) and its next inbranch
			else if ((nextInKey == 0) && (Integer.bitCount(lastInKey) > 1)) { //if parent has only one tranistion further backtracking is impossible
				lastTransition = Integer.lowestOneBit(lastInKey); //find the largest remaining in transition of this parent
				lastInKey = lastTransition ^ lastInKey;//remove this largest transition to get back to candidates' grandparent
				nextInKey = lastInKey | (lastTransition >> 1); //add the next input transition to grandparent (should always be possible)
			}
		}
		return nextInKey;
	}
	
	//identifies obviously useless places that do not need to be part of the search space
	private boolean isUseful(final int in, final int out) {
		boolean isUseful = true;
		//identify candidates inherently useless
		if ((in == 0) || (out == 0)) {//empty in or output
			isUseful = false;
		}
		if (in == out) {//pur selfloop
			isUseful = false;
		}
		if (((out & startMask) > 0) || ((in & endMask) > 0)) {//output to start or input from end
			System.out.println("Candidates: found useless place!");
			isUseful = false;
		}
/*		//indentify candidates indicated by transitions_restrictions (user-defined):
		//transition_restrictions = {inOneOutThis, inThisOutOne, inXoutY_X, inXoutY_Y};
		if ((Integer.bitCount(in) == 1) && (Integer.bitCount(out) > place_restrictions[0])) {
			isUseful = false;
		}
		if ((Integer.bitCount(out) == 1) && (Integer.bitCount(in) > place_restrictions[1])) {
			isUseful = false;
		}
		if ((Integer.bitCount(in) > place_restrictions[2]) && (Integer.bitCount(out) > place_restrictions[3])) {
			isUseful = false;
		}
*/		//update statistics and return result
		if (isUseful) {
			PlugInStatistics.instance().incNumUseFul();
		} else {
			PlugInStatistics.instance().incNumUseless();
		}
		return isUseful;
	}

	//for a given position in the given transition array return the corresponding bitmask
	private int getMask(final int pos) {
		return 1 << (transitions.length - 1 - pos);
	}

}